"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const deepmerge_1 = __importDefault(require("deepmerge"));
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const nested_error_1 = require("../error/nested-error");
const user_error_1 = require("../error/user-error");
const meta_1 = require("../meta");
const default_1 = __importDefault(require("./default"));
const severity_1 = require("./severity");
const recommended = require("./recommended");
const document = require("./document");
let rootDirCache = null;
function mergeInternal(base, rhs) {
    return deepmerge_1.default(base, rhs);
}
function loadFromFile(filename) {
    let json;
    try {
        json = require(filename);
    }
    catch (err) {
        throw new user_error_1.UserError(`Failed to read configuration from "${filename}"`, err);
    }
    for (const key of ["extends", "elements", "plugins"]) {
        if (!json[key])
            continue;
        json[key] = json[key].map((ref) => {
            return Config.expandRelative(ref, path_1.default.dirname(filename));
        });
    }
    return json;
}
class Config {
    static empty() {
        return new Config({
            extends: [],
            rules: {},
            plugins: [],
            transform: {},
        });
    }
    static fromObject(options) {
        return new Config(options);
    }
    static fromFile(filename) {
        const configdata = loadFromFile(filename);
        return new Config(configdata);
    }
    static defaultConfig() {
        return new Config(default_1.default);
    }
    constructor(options) {
        const initial = {
            extends: [],
            plugins: [],
            rules: {},
            transform: {},
        };
        this.config = mergeInternal(initial, options || {});
        this.metaTable = null;
        this.rootDir = this.findRootDir();
        this.plugins = this.loadPlugins(this.config.plugins || []);
        this.configurations = this.loadConfigurations(this.plugins);
        for (const extend of this.config.extends) {
            this.config = this.extendConfig(extend);
        }
    }
    init() {
        this.transformers = this.precompileTransformers(this.config.transform || {});
    }
    merge(rhs) {
        return new Config(mergeInternal(this.config, rhs.config));
    }
    extendConfig(entry) {
        let base;
        if (this.configurations.has(entry)) {
            base = this.configurations.get(entry);
        }
        else {
            base = Config.fromFile(entry).config;
        }
        return mergeInternal(base, this.config);
    }
    getMetaTable() {
        if (this.metaTable) {
            return this.metaTable;
        }
        const metaTable = new meta_1.MetaTable();
        const source = this.config.elements || ["html5"];
        const root = path_1.default.resolve(__dirname, "..", "..");
        for (const plugin of this.getPlugins()) {
            if (plugin.elementSchema) {
                metaTable.extendValidationSchema(plugin.elementSchema);
            }
        }
        for (const entry of source) {
            if (typeof entry !== "string") {
                metaTable.loadFromObject(entry);
                continue;
            }
            const filename = `${root}/elements/${entry}.json`;
            if (fs_1.default.existsSync(filename)) {
                metaTable.loadFromFile(filename);
                continue;
            }
            if (fs_1.default.existsSync(entry)) {
                metaTable.loadFromFile(entry);
                continue;
            }
            metaTable.loadFromObject(require(entry));
        }
        metaTable.init();
        return (this.metaTable = metaTable);
    }
    static expandRelative(src, currentPath) {
        if (src[0] === ".") {
            return path_1.default.normalize(`${currentPath}/${src}`);
        }
        return src;
    }
    get() {
        const config = Object.assign({}, this.config);
        if (config.elements) {
            config.elements = config.elements.map((cur) => cur.replace(this.rootDir, "<rootDir>"));
        }
        return config;
    }
    getRules() {
        const rules = new Map();
        for (const [ruleId, data] of Object.entries(this.config.rules || {})) {
            let options = data;
            if (!Array.isArray(options)) {
                options = [options, {}];
            }
            else if (options.length === 1) {
                options = [options[0], {}];
            }
            const severity = severity_1.parseSeverity(options[0]);
            rules.set(ruleId, [severity, options[1]]);
        }
        return rules;
    }
    getPlugins() {
        return this.plugins;
    }
    loadPlugins(plugins) {
        return plugins.map((moduleName) => {
            const plugin = require(moduleName.replace("<rootDir>", this.rootDir));
            plugin.name = moduleName;
            return plugin;
        });
    }
    loadConfigurations(plugins) {
        const configs = new Map();
        configs.set("htmlvalidate:recommended", recommended);
        configs.set("htmlvalidate:document", document);
        for (const plugin of plugins) {
            for (const [name, config] of Object.entries(plugin.configs || {})) {
                configs.set(`${plugin.name}:${name}`, new Config(config).config);
            }
        }
        return configs;
    }
    transform(filename) {
        const transformer = this.findTransformer(filename);
        if (transformer) {
            try {
                return transformer.fn(filename);
            }
            catch (err) {
                throw new nested_error_1.NestedError(`When transforming "${filename}": ${err.message}`, err);
            }
        }
        else {
            const data = fs_1.default.readFileSync(filename, { encoding: "utf8" });
            return [
                {
                    data,
                    filename,
                    line: 1,
                    column: 1,
                    originalData: data,
                },
            ];
        }
    }
    findTransformer(filename) {
        return this.transformers.find((entry) => entry.pattern.test(filename));
    }
    precompileTransformers(transform) {
        return Object.entries(transform).map(([pattern, module]) => {
            return {
                pattern: new RegExp(pattern),
                fn: require(module.replace("<rootDir>", this.rootDir)),
            };
        });
    }
    findRootDir() {
        if (rootDirCache !== null) {
            return rootDirCache;
        }
        let current = process.cwd();
        for (;;) {
            const search = path_1.default.join(current, "package.json");
            if (fs_1.default.existsSync(search)) {
                return (rootDirCache = current);
            }
            const child = current;
            current = path_1.default.dirname(current);
            if (current === child) {
                break;
            }
        }
        return (rootDirCache = process.cwd());
    }
}
exports.Config = Config;
