"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const code_frame_1 = require("@babel/code-frame");
const chalk_1 = __importDefault(require("chalk"));
const path_1 = __importDefault(require("path"));
function pluralize(word, count) {
    return count === 1 ? word : `${word}s`;
}
function formatFilePath(filePath, line, column) {
    let relPath = path_1.default.relative(process.cwd(), filePath);
    if (line && column) {
        relPath += `:${line}:${column}`;
    }
    return chalk_1.default.green(relPath);
}
function getStartLocation(message) {
    return {
        line: message.line,
        column: message.column,
    };
}
function getEndLocation(message, source) {
    let line = message.line;
    let column = message.column;
    for (let i = 0; i < message.size; i++) {
        if (source.charAt(message.offset + i) === "\n") {
            line++;
            column = 0;
        }
        else {
            column++;
        }
    }
    return { line, column };
}
function formatMessage(message, parentResult) {
    const type = message.severity === 2 ? chalk_1.default.red("error") : chalk_1.default.yellow("warning");
    const msg = `${chalk_1.default.bold(message.message.replace(/([^ ])\.$/, "$1"))}`;
    const ruleId = chalk_1.default.dim(`(${message.ruleId})`);
    const filePath = formatFilePath(parentResult.filePath, message.line, message.column);
    const sourceCode = parentResult.source;
    const firstLine = [
        `${type}:`,
        `${msg}`,
        ruleId ? `${ruleId}` : "",
        sourceCode ? `at ${filePath}:` : `at ${filePath}`,
    ]
        .filter(String)
        .join(" ");
    const result = [firstLine];
    if (sourceCode) {
        result.push(code_frame_1.codeFrameColumns(sourceCode, {
            start: getStartLocation(message),
            end: getEndLocation(message, sourceCode),
        }, { highlightCode: false }));
    }
    return result.join("\n");
}
function formatSummary(errors, warnings) {
    const summaryColor = errors > 0 ? "red" : "yellow";
    const summary = [];
    if (errors > 0) {
        summary.push(`${errors} ${pluralize("error", errors)}`);
    }
    if (warnings > 0) {
        summary.push(`${warnings} ${pluralize("warning", warnings)}`);
    }
    const output = chalk_1.default[summaryColor].bold(`${summary.join(" and ")} found.`);
    return output;
}
function codeframe(results) {
    let errors = 0;
    let warnings = 0;
    const resultsWithMessages = results.filter(result => result.messages.length > 0);
    let output = resultsWithMessages
        .reduce((resultsOutput, result) => {
        const messages = result.messages.map(message => `${formatMessage(message, result)}\n\n`);
        errors += result.errorCount;
        warnings += result.warningCount;
        return resultsOutput.concat(messages);
    }, [])
        .join("\n");
    output += "\n";
    output += formatSummary(errors, warnings);
    return errors + warnings > 0 ? output : "";
}
module.exports = codeframe;
