"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const context_1 = require("../context");
const dom_1 = require("../dom");
const event_1 = require("../event");
const lexer_1 = require("../lexer");
class Parser {
    constructor(config) {
        this.event = new event_1.EventHandler();
        this.dom = undefined;
        this.metaTable = config.getMetaTable();
    }
    parseHtml(source) {
        if (typeof source === "string") {
            source = {
                data: source,
                filename: "inline",
                line: 1,
                column: 1,
            };
        }
        this.dom = new dom_1.DOMTree({
            filename: source.filename,
            offset: 0,
            line: source.line,
            column: source.column,
        });
        this.trigger("dom:load", {
            location: null,
        });
        const lexer = new lexer_1.Lexer();
        const tokenStream = lexer.tokenize(source);
        let it = this.next(tokenStream);
        while (!it.done) {
            const token = it.value;
            switch (token.type) {
                case lexer_1.TokenType.TAG_OPEN:
                    this.consumeTag(source, token, tokenStream);
                    break;
                case lexer_1.TokenType.WHITESPACE:
                    this.trigger("whitespace", {
                        text: token.data[0],
                        location: token.location,
                    });
                    this.appendText(token.data[0], token.location);
                    break;
                case lexer_1.TokenType.DIRECTIVE:
                    this.consumeDirective(token);
                    break;
                case lexer_1.TokenType.CONDITIONAL:
                    this.trigger("conditional", {
                        condition: token.data[1],
                        location: token.location,
                    });
                    break;
                case lexer_1.TokenType.DOCTYPE_OPEN:
                    this.consumeDoctype(token, tokenStream);
                    break;
                case lexer_1.TokenType.TEXT:
                    this.appendText(token.data, token.location);
                    break;
                case lexer_1.TokenType.EOF:
                    this.closeTree(source, token.location);
                    break;
            }
            it = this.next(tokenStream);
        }
        this.dom.resolveMeta(this.metaTable);
        this.trigger("dom:ready", {
            document: this.dom,
            location: null,
        });
        return this.dom;
    }
    closeOptional(token) {
        const active = this.dom.getActive();
        if (!(active.meta && active.meta.implicitClosed)) {
            return false;
        }
        const tagName = token.data[2];
        const open = !token.data[1];
        const meta = active.meta.implicitClosed;
        if (open) {
            return meta.indexOf(tagName) >= 0;
        }
        else {
            if (active.is(tagName)) {
                return false;
            }
            return active.parent.is(tagName) && meta.indexOf(active.tagName) >= 0;
        }
    }
    consumeTag(source, startToken, tokenStream) {
        const tokens = Array.from(this.consumeUntil(tokenStream, lexer_1.TokenType.TAG_CLOSE));
        const endToken = tokens.slice(-1)[0];
        const closeOptional = this.closeOptional(startToken);
        const parent = closeOptional
            ? this.dom.getActive().parent
            : this.dom.getActive();
        const node = dom_1.HtmlElement.fromTokens(startToken, endToken, parent, this.metaTable);
        const open = !startToken.data[1];
        const close = !open || node.closed !== dom_1.NodeClosed.Open;
        const foreign = node.meta && node.meta.foreign;
        if (closeOptional) {
            const active = this.dom.getActive();
            active.closed = dom_1.NodeClosed.ImplicitClosed;
            this.closeElement(source, node, active, startToken.location);
            this.dom.popActive();
        }
        if (open) {
            this.dom.pushActive(node);
            this.trigger("tag:open", {
                target: node,
                location: startToken.location,
            });
        }
        for (let i = 0; i < tokens.length; i++) {
            const token = tokens[i];
            switch (token.type) {
                case lexer_1.TokenType.WHITESPACE:
                    break;
                case lexer_1.TokenType.ATTR_NAME:
                    this.consumeAttribute(source, node, token, tokens[i + 1]);
                    break;
            }
        }
        if (close) {
            const active = this.dom.getActive();
            if (!open) {
                node.closed = dom_1.NodeClosed.EndTag;
            }
            this.closeElement(source, node, active, endToken.location);
            const voidClosed = !open && node.voidElement;
            if (!voidClosed) {
                this.dom.popActive();
            }
        }
        else if (foreign) {
            this.discardForeignBody(source, node.tagName, tokenStream);
        }
    }
    closeElement(source, node, active, location) {
        if (source.hooks && source.hooks.processElement) {
            const processElement = source.hooks.processElement;
            processElement(active);
        }
        this.trigger("tag:close", {
            target: node,
            previous: active,
            location,
        });
        if (active && !active.isRootElement()) {
            this.trigger("element:ready", {
                target: active,
                location: active.location,
            });
        }
    }
    discardForeignBody(source, foreignTagName, tokenStream) {
        let nested = 1;
        let startToken;
        let endToken;
        do {
            const tokens = Array.from(this.consumeUntil(tokenStream, lexer_1.TokenType.TAG_OPEN));
            const [last] = tokens.slice(-1);
            const [, tagClosed, tagName] = last.data;
            if (tagName !== foreignTagName) {
                continue;
            }
            const endTokens = Array.from(this.consumeUntil(tokenStream, lexer_1.TokenType.TAG_CLOSE));
            endToken = endTokens.slice(-1)[0];
            const selfClosed = endToken.data[0] === "/>";
            if (tagClosed) {
                startToken = last;
                nested--;
            }
            else if (!selfClosed) {
                nested++;
            }
        } while (nested > 0);
        const active = this.dom.getActive();
        const node = dom_1.HtmlElement.fromTokens(startToken, endToken, active, this.metaTable);
        this.closeElement(source, node, active, endToken.location);
        this.dom.popActive();
    }
    consumeAttribute(source, node, token, next) {
        const keyLocation = token.location;
        const valueLocation = this.getAttributeValueLocation(next);
        const haveValue = next && next.type === lexer_1.TokenType.ATTR_VALUE;
        const attrData = {
            key: token.data[1],
        };
        if (haveValue) {
            attrData.value = next.data[1];
            attrData.quote = next.data[2];
        }
        let processAttribute = (attr) => [
            attr,
        ];
        if (source.hooks && source.hooks.processAttribute) {
            processAttribute = source.hooks.processAttribute;
        }
        let iterator;
        const legacy = processAttribute(attrData);
        if (typeof legacy[Symbol.iterator] !== "function") {
            iterator = [attrData];
        }
        else {
            iterator = legacy;
        }
        for (const attr of iterator) {
            this.trigger("attr", {
                target: node,
                key: attr.key,
                value: attr.value,
                quote: attr.quote,
                originalAttribute: attr.originalAttribute,
                location: keyLocation,
                valueLocation,
            });
            node.setAttribute(attr.key, attr.value, keyLocation, valueLocation, attr.originalAttribute);
        }
    }
    getAttributeValueLocation(token) {
        if (!token || token.type !== lexer_1.TokenType.ATTR_VALUE || token.data[1] === "") {
            return null;
        }
        const quote = token.data[2];
        if (quote) {
            return context_1.sliceLocation(token.location, 2, -1);
        }
        else {
            return context_1.sliceLocation(token.location, 1);
        }
    }
    consumeDirective(token) {
        const directive = token.data[1];
        const match = directive.match(/^([a-zA-Z0-9-]+)\s*(.*?)(?:\s*:\s*(.*))?$/);
        if (!match) {
            throw new Error(`Failed to parse directive "${directive}"`);
        }
        const [, action, data, comment] = match;
        this.trigger("directive", {
            action,
            data,
            comment: comment || "",
            location: token.location,
        });
    }
    consumeDoctype(startToken, tokenStream) {
        const tokens = Array.from(this.consumeUntil(tokenStream, lexer_1.TokenType.DOCTYPE_CLOSE));
        const doctype = tokens[0];
        const value = doctype.data[0];
        this.dom.doctype = value;
        this.trigger("doctype", {
            value,
            location: startToken.location,
        });
    }
    *consumeUntil(tokenStream, search) {
        let it = this.next(tokenStream);
        while (!it.done) {
            const token = it.value;
            yield token;
            if (token.type === search)
                return;
            it = this.next(tokenStream);
        }
        throw Error("stream ended before consumeUntil finished");
    }
    next(tokenStream) {
        return tokenStream.next();
    }
    on(event, listener) {
        return this.event.on(event, listener);
    }
    once(event, listener) {
        return this.event.once(event, listener);
    }
    defer(cb) {
        this.event.once("*", cb);
    }
    trigger(event, data) {
        if (typeof data.location === "undefined") {
            throw Error("Triggered event must contain location");
        }
        this.event.trigger(event, data);
    }
    getEventHandler() {
        return this.event;
    }
    appendText(text, location) {
        this.dom.getActive().appendText(text, location);
    }
    closeTree(source, location) {
        let active;
        while ((active = this.dom.getActive()) && !active.isRootElement()) {
            this.closeElement(source, null, active, location);
            this.dom.popActive();
        }
    }
}
exports.Parser = Parser;
