"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const context_1 = require("../context");
const dom_1 = require("../dom");
const rule_1 = require("../rule");
const defaults = {
    relaxed: false,
};
const textRegexp = /([<>]|&(?![a-zA-Z0-9#]+;))/g;
const unquotedAttrRegexp = /([<>"'=`]|&(?![a-zA-Z0-9#]+;))/g;
const replacementTable = {
    '"': "&quot;",
    "&": "&amp;",
    "'": "&apos;",
    "<": "&lt;",
    "=": "&equals;",
    ">": "&gt;",
    "`": "&grave;",
};
class NoRawCharacters extends rule_1.Rule {
    constructor(options) {
        super(Object.assign({}, defaults, options));
        this.relaxed = this.options.relaxed;
    }
    documentation() {
        return {
            description: `Some characters such as \`<\`, \`>\` and \`&\` hold special meaning in HTML and must be escaped using a character reference (html entity).`,
            url: rule_1.ruleDocumentationUrl(__filename),
        };
    }
    setup() {
        this.on("element:ready", (event) => {
            const node = event.target;
            for (const child of node.childNodes) {
                if (child.nodeType !== dom_1.NodeType.TEXT_NODE) {
                    continue;
                }
                this.findRawChars(child.textContent, child.location, textRegexp);
            }
        });
        this.on("attr", (event) => {
            if (!event.value) {
                return;
            }
            if (event.quote) {
                return;
            }
            this.findRawChars(event.value.toString(), event.valueLocation, unquotedAttrRegexp);
        });
    }
    findRawChars(text, location, regexp) {
        let match;
        do {
            match = regexp.exec(text);
            if (match) {
                const char = match[0];
                if (this.relaxed && char === "&") {
                    continue;
                }
                const replacement = replacementTable[char];
                const charLocation = context_1.sliceLocation(location, match.index, match.index + 1);
                this.report(null, `Raw "${char}" must be encoded as "${replacement}"`, charLocation);
            }
        } while (match);
    }
}
module.exports = NoRawCharacters;
